/*
 * flicksoup.h -- Main include file
 *
 * Copyright (C) 2010-2012 Mario Sanchez Prada
 * Authors: Mario Sanchez Prada <msanchez@gnome.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Lesser Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>
 *
 */

#ifndef FLICKSOUP_H
#define FLICKSOUP_H

G_BEGIN_DECLS

#include <glib.h>
#include <glib-object.h>

#include <flicksoup/fsp-session.h>
#include <flicksoup/fsp-data.h>
#include <flicksoup/fsp-error.h>

G_END_DECLS

#endif
